% :- module( blankets, [blankets/2] ).

:- ensure_loaded( markov_blanket ).	% mbs/2, kv_compose_pivot_v/3.

/* 
main( Inp ) :-
	( Inp = [InFile|Tail] ->
		( Tail = [OutFile] ->
			true
			;
			OutFile = user_output
		)
		;
		InFile = Inp
	),
	main( InFile, OutFile ).
*/

blankets( InFile, OutFile ) :-
	Pivot = [1-0,2-0,3-0,4-0,5-0,6-0,7-0,8-0],
	kv_compose_pivot_v( [1,2,3,4,5,6,7,8], Pivot, Empty ),
	open( InFile, read, Stream ),
	count( Stream, 0, Empty, Total, Counts ),
	close( Stream ),
	current_output( OldOut ),
	open( OutFile, write, Out ),
	set_output( Out ),
	output_counts( Counts, Total ),
	set_output( OldOut ).

count( Stream, InTot, Acc, OutTot, Counts ) :-
	read( Stream, Term ),
	( Term = end_of_file -> 
		Counts = Acc,
		OutTot is InTot
		;
		( Term = c( Bn ) ->
			NxTot is InTot + 1,
			mbs( Bn, MBs ),
			% write( mbs( Bn, MBs ) ), nl,
			update_counts( MBs, Acc, NxAcc )
			;
			NxAcc = Acc,
			NxTot is InTot
		),
		count( Stream, NxTot, NxAcc, OutTot, Counts )
	).

output_counts( Counts, Total ) :-
	write( 'blanket_counts( [' ), nl,
	output_counts_list( Counts, Total ),
	write( '] ).' ).

output_counts_list( [], _Total ).
output_counts_list( [Hk-Hlist|T], Total ) :-
	tab( 1 ), write( Hk ), write( ' - [' ), nl,
	output_blanket_counts_list( Hlist, Total ),
	tab( 1 ), write( '    ]' ),
	( T == [] -> nl ; write( ',' ), nl ),
	output_counts_list( T, Total ).
	
output_blanket_counts_list( [], _Total ).
output_blanket_counts_list( [Hk-Hv|T], Total ) :-
	Quot is Hv / Total,
	tab( 3 ), write( Hk-Quot ),
	( T = [] -> nl ; write( ',' ), nl ),
	output_blanket_counts_list( T, Total ).

update_counts( [], Out, Out ).
update_counts( [Vx-VxBlanket|T], In, Out ) :-
	kv_swap_v( In, Vx, OldVal, NwVal, Mid ),
	kv_increment_vs_of_ks( VxBlanket, OldVal, NwVal ),
	update_counts( T, Mid, Out ).

kv_swap_v( [Hk-Hv|T], Key, OldVal, NewVal, Swap ) :-
	( Hk < Key -> 
		Swap = [Hk-Hv|Tswap],
		kv_swap_v( T, Key, OldVal, NewVal, Tswap )
		;
		( Hk =:= Key ->
			OldVal = Hv, 
			Swap = [Hk-NewVal|T]
			;
			write( error_in_kv_swap_v ), nl,
			abort
		)
	).

kv_increment_vs_of_ks( [], KVs, KVs ).
kv_increment_vs_of_ks( [Hk|Keys], KVs, NewKVs ) :-
	kv_increment_vs_of_key_w_rem( KVs, Hk, RemKVs, NewKVs, NewTail ),
	kv_increment_vs_of_ks( Keys, RemKVs, NewTail ).

kv_increment_vs_of_key_w_rem( [Hk-Hv|Tkvs], Key, RemKVs, NewKVs, NewTail ) :-
	( Hk < Key ->
		NewKVs = [Hk-Hv|TnewKVs],
		kv_increment_vs_of_key_w_rem( Tkvs, Key, RemKVs, TnewKVs, NewTail )
		;
		( Hk =:= Key ->
			RemKVs = Tkvs,
			NewHv is Hv + 1,
			NewKVs = [Hk-NewHv|NewTail]
			;
			write( error_in_kv_increment_vs_of_key_w_rem ), nl,
			abort
		)
	).
