:- ensure_loaded( '../src/init_lib' ).

:- ensure_loaded( library(file_to_list_of_lines) ).
:- ensure_loaded( library(breaks_list) ).
:- ensure_loaded( disp_bn ).
:- ensure_loaded( library(lists) ). 				% append/3,reverse/2.
:- pl( sicstus(_S), ensure_loaded(library(number_atom)) ). % /2.
:- pl( yap(_Y), ensure_loaded(library(at_end_of_line)) ).


% Asia BN
default_bn( [1-[],2-[1],3-[2,5],4-[],5-[4],6-[4],7-[3],8-[3,6]] ).

% Magenta arrows: ones that dont exist in generator.
% Yellow arrows: ones that dont exist in guessed.
% Red arrows: ones that exist in both, the generator 
%               and guessed BNs.
% disp_cut_off( 0.95, '/cdrom/uai01_results/pV28/by_ext/counts/auto_pV28_2pun_3h25_half_i500K_s242.counts' ).
disp_cut_off( [CutOffEither,File] ) :-
	disp_cut_off( CutOffEither, File ).
disp_cut_off( CutOffEither, File ) :-
	( number(CutOffEither) ->
		CutOff = CutOffEither
		;
		number_atom( CutOff, CutOffEither )
	),
	write( cutoff(CutOff) ), nl,
	file_to_list_of_lines( File, Lines ),
	built_cut_bn( Lines, CutOff, Bn ),
	write( bn(Bn) ), nl,
	default_bn( Def ),
	disp_bn( Bn, Def ).

built_cut_bn( Lines, CutOff, Bn ) :-
	( has_node_with_lines( Lines, Node, NodeLines, RestLines ) ->
		node_lines_to_parents( NodeLines, CutOff, [], Parents ),
		% write( parents(Parents) ), nl,
		Bn = [Node-Parents|TBn],
		built_cut_bn( RestLines, CutOff, TBn )
		;
		Bn = []
	).

has_node_with_lines( Lines, Node, NodeLines, RestLines ) :-
	TpPred = append("Parent counts for ",NodeCsCol,E),
	breaks_list(Lines,TpPred,E,_LeftLns,_The,RightLns),
	append( NodeCs, [0':|_Rest], NodeCsCol ),
	number_codes( Node, NodeCs ),
	BtPred = append("Parent counts for ",_OthCs,BtE),
	( breaks_list(RightLns,BtPred,BtE,NodeLines,TheLn,RRemLns) ->
		RestLines = [TheLn|RRemLns]
		;
		RestLines = [],
		NodeLines = RightLns
	).

node_lines_to_parents( [], _CutOff, Acc, Parents ) :-
	reverse( Acc, Parents ).
node_lines_to_parents( [H|T], CutOff, AccIn, Parents ) :-
	node_line_to_parent( H, CutOff,  AccIn, AccMid ),
	node_lines_to_parents( T, CutOff, AccMid, Parents ).

node_line_to_parent( [], _CutOff,  Acc, Acc ) :- !.
node_line_to_parent( Line, CutOff,  AccIn, AccOut ) :-
	( append( Left, [0' |Right], Line ) ->
		number_codes( Prob, Right ),
		( Prob > CutOff  ->
			number_codes( Parent, Left ),
			AccOut = [Parent|AccIn]
			;
			AccOut = AccIn
		)
		;
		write( node_line_to_parent(unrecognisided_line_format(Line)) ), nl,
		abort
	).
