:- ensure_loaded( library(ordsets) ).		% ord_add_element/3, ord_union/2, ord_union/2.
:- ensure_loaded( library(ugraphs) ).		% vertices/2.

% mbs( [1-[],2-[1],3-[2,5],4-[],5-[4],6-[4],7-[3],8-[3,6]], B ).

mbs( Bn, Mbs ) :-
	vertices( Bn, Vs ),
	kv_compose_pivot_v( Vs, [], Kvs ),
	parental_graph( Bn, Kvs, Pbn ),
	mbs( Vs, Bn, Pbn, Mbs ).

mbs( [], _Bn, _Pbn, [] ).
mbs( [V|Vs], Bn, Pbn, [V-Vblanket|Mbs] ) :-
	neighbours( V, Bn, Parents ),
	neighbours( V, Pbn, Children ),
	all_neighbours( Children, Bn, [], AllCoParents ),
	ord_del_element( AllCoParents, V, CoParents ),
	ord_union( [Children,Parents,CoParents], Vblanket ),
	mbs( Vs, Bn, Pbn, Mbs ).

all_neighbours( [], _Graph, All, All ).
all_neighbours( [V|Vs], Graph, Acc, All ) :-
	neighbours( V, Graph, These ),
	ord_union( These, Acc, NxAcc ),
	all_neighbours( Vs, Graph, NxAcc, All ).

parental_graph( [], Parental, Parental ).
parental_graph( [Vx-Children|Graph], AccParental, Parental ) :-
	update_parental( Children, Vx, AccParental, NxParental ),
	parental_graph( Graph, NxParental, Parental ).

update_parental( [], _Vx, Parental, Parental ).
update_parental( [H|T], Vx, AccParental, Parental ) :-
	kvs_add_elem_to_v( AccParental, H, Vx, NxParental ),
	update_parental( T, Vx, NxParental, Parental ).

kvs_add_elem_to_v( [], Key, Value, [Key-[Value]] ).
kvs_add_elem_to_v( [Hk-Hv|Tkv], Key, Value, NwKv ) :-
	( Hk < Key ->
		NwKv = [Hk-Hv|TnwKv],
		kvs_add_elem_to_v( Tkv, Key, Value, TnwKv )
		;
		( Hk =:= Key  ->
			ord_add_element( Hv, Value, NwV ),
			NwKv = [Hk-NwV|Tkv]
			;
			NwKv = [Key-[Value],Hk-Hv|Tkv]
		)
	).

kv_compose_pivot_v( [], _Pivot, [] ).
kv_compose_pivot_v( [H|T], Pivot, [H-Pivot|Tkv] ) :-
	kv_compose_pivot_v( T, Pivot, Tkv ).
