:- ensure_loaded( 'src/init_lib' ).

:- ensure_loaded( library(pl) ). 		% pl/2.

:- pl( yap(_Y), yap_flag(update_semantics,logical) ).

:- ensure_loaded( 'src/all_dynamic' ).
:- ensure_loaded( 'src/ad_to_slp' ).
:- ensure_loaded( 'src/kernel' ).
% :- ensure_loaded( 'src/kernel_op' ).
% :- ensure_loaded( 'src/kernel_vp' ).

:- ensure_loaded( library(lss) ).					% lss/0.
:- ensure_loaded( library(flatoms) ).				% /2.
:- ensure_loaded( library(unique_filename) ).		% /2.
:- ensure_loaded( library(random) ).
:- ensure_loaded( library(copy_stream_on_stream) ).
:- ensure_loaded( library(dbg) ).		% expects debug_these/1, provides dbg/1,2.
% :- ensure_loaded( library(esr) ).		% /1.

:- pl( sicstus(_S), ensure_loaded( library(number_atom) ) ).  		% /2.

dbg_these( [24,28,29] ).   	% for long runs. c/1, and b/1.

data_format(data(_,_,_,_,_,_,_,_)).
:- ensure_loaded( 'src/jc_llhood_ratio_fa' ).
% :- bb_put( sr, off ). % To print on screen.

% Args = 
% File, Stats, KernelId, BnFile, PofBacktracking, Order, ConstraintsInBN,
% Repeats, DataFile.  Sn, Seeds.
mcmcms( Args ) :-
	esr(prolog_flag( user_error, OldError )),
	esr(current_output( OldOutput )),
	write( args( Args ) ), nl,
	Args = [F,S,K,B,P,O,R,D,S1,S2,S3],
	number_atom_list( [Np,Nr,Ns1,Ns2,Ns3], [P,R,S1,S2,S3] ),
	( bb_get(run_pl_directory,RunDir) -> true ; RunDir = '.' ),
	% flatoms( [RunDir,'/src/kernel_',K], Kfile ),
	% pl( sicstus(_S), (ensure_loaded(Kfile),ensure_loaded(D)) ),
	% pl( yap(_Y), (reconsult(Kfile),reconsult(D)) ),
	pl( sicstus(_S), (ensure_loaded(D)) ),
	pl( yap(_Y), (reconsult(D)) ),
	% write( reconsulting_ok ), nl,  % Yap 4,5,1 fails somewhere here.
	esr(open( S, write, Stats )),
	esr(open( F, write, Results )),
	esr(set_prolog_flag( user_error, Stats )),
	% set_output( Results ),
	print_args( Args ),
	setrand( rand(Ns1,Ns2,Ns3) ),
	bn_file_abs( B, AbsB ),
	all_dynamic( AbsB ),
	TmpLoad = googles_tmp,
	ad_to_slp( [pred_type(_A,_B),s_label(_C,_D)], dont_delete, TmpLoad ),
	( O == or -> 
		Nodes = [1,4,2,5,6,3,7,8]
		;
		Nodes = [1,2,3,4,5,6,7,8]
	),
	write( user_error, nodes(Nodes) ), nl( user_error ),
	( var(Stats) -> WSon = user ; WSon = Stats ),
	( main_out( WSon, Results, K, Nodes, Nr, Np, TmpLoad ) -> true;
						write( user_error, finished_with_failure ), nl
	),
	esr(set_output( OldOutput )),
	esr(close( Results )),
	esr(set_prolog_flag( user_error, OldError )),
	esr(close( Stats )).

main_out( Stats, Results, Kid, Nodes, Nr, Np ) :-
	main_out( Stats, Results, Kid, Nodes, Nr, Np, none__ ).

main_out( Stats, Results, Kid, Nodes, Nr, Np, TmpLoad ) :-
	esr(set_output( Stats )),
	prolog_flag( version, ThisPlEngine ),
	write( on(ThisPlEngine) ), nl,
	% lss, this doesnt seem to work on sicstus 3.10.1
	( TmpLoad == none__ ->
		true
		;
		open( TmpLoad, read, ReadFrom  ),
		copy_stream_on_stream( ReadFrom, Stats ),
		delete_file_if( TmpLoad )
	),
	% end of patch 
	main_statistics( StartAt ),
	esr(set_output( Results )),
	kernel( Nodes, Nr, Np, Kid, _M ),
	esr(set_output( Stats )),
	main_statistics( EndWith ),
	Total is (EndWith - StartAt) / 1000,
	write( ((EndWith - StartAt) / 1000) ), write( ' = ' ), nl,
	write( Total ), nl.

print_args( [F,S,K,B,P,O,R,D,S1,S2,S3] ) :-
	write( user_error, results_file(F) ), nl(user_error ),
	write( user_error, stats_file(S) ), nl(user_error ),
	write( user_error, kernel_file(K) ), nl(user_error ),
	write( user_error, bn_file(B) ), nl(user_error ),
	write( user_error, backtrack_probability(P) ), nl(user_error ),
	write( user_error, order(O) ), nl(user_error ),
	% write( user_error, constrainrs(C) ), nl(user_error ),
	write( user_error, repeats(R) ), nl(user_error ),
	write( user_error, data_file(D) ), nl(user_error ),
	write( user_error, seed1(S1) ), nl(user_error ),
	write( user_error, seed2(S2) ), nl(user_error ),
	write( user_error, seed3(S3) ), nl(user_error ).

number_atom_list( [], [] ).
number_atom_list( [H|T], [Nh|Nt] ) :-
	number_atom( H, Nh ),
	number_atom_list( T, Nt ).

bn_file_abs( Bn, AbsBn ) :-
	( pl(yap(_Y)) -> 
		library_directory( Lib ),
		flatoms( [Lib,'/',Bn], AbsBn ),
		file_exists( AbsBn ), 
		!
		;
		AbsBn = library( Bn )
	).

main_statistics( Sum ) :-
	pl( Pl ),
	( Pl = sicstus(_) ->
		statistics( atom_garbage_collection, [_Nofag,_Ab,OvAt] )
		;
		statistics( cputime, [_OvCt,CpuTime] ),
		write( yap_cputime( CpuTime ) ), nl,
		OvAt is 0
	),
	statistics( runtime, [OvRt,_RlRt] ),
	statistics( garbage_collection, [_Nofgc,_Bfrd,OvGt] ),
	statistics( stack_shifts, [_Nofls,_Nofts,OvSt] ),
	Sum is (OvRt + OvGt + OvSt + OvAt),
	statistics.
