% run/1 uses the run_data/8, residing in file RunsFile, to run a number
% of experiments described in the RunsFile.
% All resulting files are placed in directory ResDir.
% run/1 takes a list as its only argument and the terms recognised
% within this list are 
%   id(Id)  there can be multiple instances of those
%   runs_file(RunsFile) as above
%   results_dir(ResDir) as above
%   display(DisplayPred) pointing to a rule deciding if a choice should be 
%     displayed using graphviz.
%
% The following comments detail what the args in run_data/9
% (also to be found in runs/simple_test.pl).
%
% run_data( Id, Kernel, P, Ord, Cnsts, Repeats, Data, RndSeed, Exts ).
% Id, for the experiment, normally an integer.
% Kernel = which kernel to use (in {vp,op}). 
	% vp = variable (1-(1/(2**(n-1)))); op = original (unchanging p)
% P = probability of backtracking parameter(number)
	% for Kernel, vp : P is an integer (typically depending on depth of tree
	% 			here we mostly used 28)
	%		    op : 0 =< P =< 1 (typically values around 1).
% Ord in {o,u}
% Constraints, like '', '2p_3h25','3p' (basically a way to select a particular 
%              bn generator file from slp/
% Repeats, number of repetitions to perform.
% Data in {half,all}, possible to extend. DataFile = data/jc_asia_data_$Data
% RndSeed in 1..1000 as in file auxil/seeds.
% Exts in {all,none,List} List may contain {blankets,counts,visits,graphs}
% 	note : (a) `visits' also generates stays, (b) `graphs' produces same files
% 	as visits, but also produces asynchronous graphs on screen.
% 	all expands to [blankets,counts,visits]
% 
% Slp = slp/bn_O_Cnsts

:- ensure_loaded( 'src/init_lib' ).		% /2.
:- ensure_loaded( library(flatoms) ).		% /2.
:- ensure_loaded( library(sys_info) ).		% /1.
:- ensure_loaded( 'auxil/seeds' ). 		% seed/2.
:- ensure_loaded( library(system) ).		
	% file_exists/1, make_directory/1.
:- ensure_loaded( library(lists) ).		% memberchk/2, reverse/2, select/3.
:- ensure_loaded( mcmcms ).		

:- ensure_loaded( 'auxil/create_visits' ).		
:- ensure_loaded( 'auxil/create_blankets' ).		
:- ensure_loaded( 'auxil/create_graphs' ).		
:- ensure_loaded( library(esr) ).		% /1.
:- ensure_loaded( 'auxil/disp_bn' ).		% not always needed though.

:- prolog_load_context(directory,Dir), bb_put( run_pl_directory, Dir ).
:- pl( sicstus(_S), ensure_loaded(library(number_atom)) ). % /2.

% :- bb_put( sr, off ). % To print on screen. also known in mcmcms.

/* 
run( [] ).
equivalent to
run( [runs_file('runs/simple'),results_dir(test)] ).
the following is the simplest test possible.
run( [id(1),runs_file('runs/tiny'),results_dir(test)] ).
run( [id(1),runs_file('runs/seeds_effect_lnai_8pu'),results_dir(test)] ).
Following for displaying bns.
run( [display(next_display_at), id(2),runs_file('runs/small'),results_dir(test)] ).
As above, but with second iterations models being identical :
run( [display(next_display_at), id(1),runs_file('runs/small'),results_dir(test)] ).
run( [display(every_single_one_of_them),id(1)] ). obsolete ?
*/

run :- run( [] ).

run( InArgs ) :-
	( select( display(PredName), InArgs, Args ) ->
		Call =.. [PredName,Iter],
		current_output( Out ),
		bb_put( display_bn_at, (Iter,Call,wait_for_return_or_integer(Out,Iter),Out) ),
		bb_put( next_display_iteration, 1 ),
		!
		;
		Args = InArgs
	),
	( memberchk(runs_file(RunsFile),Args) -> true ; RunsFile = 'runs/simple' ),
	locate_source_file( RunsFile, run_pl_directory, RunsSourceFile ),
	bb_get( run_pl_directory, RunDir ),
	( memberchk(results_dir(ResDir),Args) -> true ;
		atom_concat( RunDir, '/test', ResDir )
		),
	( memberchk(id(Id),Args) -> 
		member( id(Id), Args ) ; true ),
	( file_exists(ResDir) -> true ;
		make_directory( ResDir )
	),
	flatoms( [ResDir,'/std_out_n_err'], Err ),
	open( Err, append, Estream ),
	current_output( OldOutStream ),
	bb_put( old_out_stream, OldOutStream ),
	prolog_flag( user_error, OldErrStream ), % do this without esr/1
		% yap uses immediate update semantics for prolog_flag.
	bb_put( old_err_stream, OldErrStream ),
	bb_put( err_stream, Estream ),
	compile( RunsSourceFile ),
	run_data( Id, Kernel, P, Ord, Cnsts, Repeats, Data, RndS, Exts ),
	esr( set_output( Estream ) ),
	esr( set_prolog_flag( user_error, Estream ) ),
	number_atom( RndS, Sat ),
	number_atom( P, Pat ),
	write(number_atom( P, Pat )), nl,
	run_filename( Kernel, Pat, Ord, Cnsts, Repeats, Data, Sat, Filename ),
	write( run_filename( Kernel, Pat, Ord, Cnsts, Repeats, Data, Sat, Filename ) ), nl,
	seed_id_to_seed_atoms( RndS, S1, S2, S3 ),
	flatoms( [ResDir,'/',Filename], DirF ),
	number_atom( Repeats, Rat ),
	% flatoms( ['src/kernel_',Kernel], Kfile ),
	% bb_get( run_pl_directory, RunDir ),
	flatoms( [RunDir,'/data/jc_asia_data_',Data], Dfile ),
	bn_filename( Ord, Cnsts, BnF ),
	flatoms( [DirF,'.stats'], Sfile ),
	mcmcms( [DirF, Sfile, Kernel, BnF, Pat, Ord, Rat, Dfile, S1, S2, S3] ),
	unfold_extensions( Exts, Unfolds ),
	exec_extensions( Unfolds, DirF ),
	flatoms( ['gzip -f ',DirF], Gzip ),
	esr( sys_info( Gzip ) ),
	fail.
run( _ ) :-
	bb_get( old_out_stream, OldOutStream ),
	set_output( OldOutStream ),
	bb_get( old_err_stream, OldErrStream ),
	set_prolog_flag( user_error, OldErrStream ),
	bb_get( err_stream, ErrStream ),
	close( ErrStream ).
	
logical_prolog_flag( A, B ) :-
	atomic( A ),
	prolog_flag( A, B ),
	!.

exec_extensions( [], _DirF ).
exec_extensions( [H|T], DirF ) :-
	exec_extension( H, DirF ),
	exec_extensions( T, DirF ).

exec_extension( counts, DirF ) :-
	!,
	bb_get( run_pl_directory, RunDir ),
	flatoms( [RunDir,'/auxil/jc_count_parents.perl',' ',DirF, '> ',DirF,'.',counts,
				' 2>> /dev/null'],
				ExecH ),
	% write( std_exec(ExecH) ), nl,
	% exec( ExecH, [std,std,std], _Pid ).
	sys_info( ExecH ).
exec_extension( H, DirF ) :-
	( memberchk(H,[blankets,graphs,visits]) -> 
		flatoms( [DirF,'.',H],ExtDf ),
		Goal =.. [H,DirF,ExtDf],
		( call( Goal ) -> 
			write( ext_goal_succ( Goal ) ), nl
			;
			write( ext_goal_fail( Goal ) ), nl
		)
		;
		write( 'Warning: unrecognised extension, ' ), 
		write( H ), write( '.' ), nl
	).

/* 
exec_extensions( Exts, DirF ) :-
	write( extensions(Exts) ), nl,
	pl( Pl ),
	Pl =.. [Name|_Args],
	flatoms( ['auxil/upsh_',Name,' '], Upsh ),
	exec_extensions( Exts, Upsh, DirF ).
	% atom_codes( Name, NameCs ),

exec_extensions( [], _Upsh, _DirF ).
exec_extensions( [H|T], Upsh, DirF ) :-
	( H == counts ->
		flatoms( ['auxil/create_counts',' ',DirF,' ',DirF,'.',H], ExecH )
		;
		flatoms( [Upsh,' auxil/create_',H,' ',DirF,' ',DirF,'.',H], ExecH )
	),
	sys_info( ExecH ),
	exec_extensions( T, Upsh, DirF ).
	*/

bn_filename( Ord, Cnsts, BnFile ) :-
	( Cnsts = '' -> 
		flatoms( ['../../slp/bn_',Ord,'.slp'], BnFile )
		;
		flatoms( ['../../slp/bn_',Ord,'_',Cnsts,'.slp'], BnFile )
	).

seed_id_to_seed_atoms( Rnd, S1, S2, S3 ) :-
	seed( Rnd, rand(X1,X2,X3) ),
	number_atom( X1, S1 ),
	number_atom( X2, S2 ),
	number_atom( X3, S3 ).

run_filename( Kernel, P, Ord, Cnsts, Repeats, Data, S, Filename ) :-
	repeats_abbreviate( Repeats, Abrv ),
	atom_codes( P, Pcs ),
	delete_trailing( Pcs, 0'0, ClPcs ),
	atom_codes( ClP, ClPcs ),
	flatoms( ['tr_',Kernel,ClP,'_',Ord,'_',Cnsts,'_i',Abrv,'_',Data,'_s',S], Filename ).

repeats_abbreviate( Repeats, Abrv ) :-
	M is Repeats / 1000000,
	( M >= 1 ->
		number_codes( M, Mcs ),
		delete_trailing( Mcs, 0'0, NzMcs ),
		append( NzMcs, "M", AbrvCs ),
		atom_codes( Abrv, AbrvCs )
		;
		K is Repeats // 1000,
		( K > 0 -> 
			number_codes( K, Kcs ),
			append( Kcs, "K", AbrvCs ),
			atom_codes( Abrv, AbrvCs )
			;
			number_codes( Repeats, RepeatsCs ),
			atom_codes( Abrv, RepeatsCs )
		)
	).

unfold_extensions( Exts, Unfolds ) :-
	( Exts == all ->
		Unfolds = [blankets,visits,counts]
		;
		( Exts == none ->
			Unfolds = []
			;
			Unfolds = Exts
		)
	).

delete_trailing( List, NoTrail, CleanList ) :-
	reverse( List, RevList ),
	delete_leading( RevList, NoTrail, RevCleanList ),
	reverse( RevCleanList, CleanList ).

delete_leading( List, Elem, Clean ) :-
	( List = [H|T] ->
		( H \= Elem -> 
			Clean = List
			;
			delete_leading( T, Elem, Clean )
		)
		;
		Clean = List
	).

%
wait_for_return( Out ) :-
	write( Out, ' press return to continue ' ), get_code( _C ), nl(Out).

every_single_one_of_them( _ ).

next_display_at( Iter ) :-
	% bb_get( last_display_iteration, Last ),
	bb_get( next_display_iteration, Next ),
	Iter >= Next.

wait_for_return_or_integer( Out, Iter ) :-
	write( Out, ' press return or give next iteration to display (integer), to continue : ' ),
	get_code( C ), nl(Out),
	( (C >= 0'0, C =< 0'9) -> 
		get_integer_from_get_codes( [C], Integer )
		;
		Integer is 1,
		consume_codes_to_return( C )
	),
	!,
	NxDisplay is Iter + Integer,
	write( Out, nx_disp_at(NxDisplay) ), nl( Out ),
	bb_put( next_display_iteration, NxDisplay ).

get_integer_from_get_codes( Acc, Integer ) :-
	get_code( C ),
	( (C >= 0'0, C =< 0'9) -> 
		get_integer_from_get_codes( [C|Acc], Integer )
		;
		reverse( Acc, IntCs ),
		number_codes( Integer, IntCs ),
		( Integer =:= 0 -> halt ; true ),
		consume_codes_to_return( C )
	).

consume_codes_to_return( 10 ) :- !.
consume_codes_to_return( _ ) :-
	get_code( C ),
	consume_codes_to_return( C ).

locate_source_file( File, BBFlag, FileLoc ) :-
	pl( sicstus(_S),
		absolute_file_name( File, Source, [file_type(source)] ) ),
	pl( yap(_Y),
		absolute_file_name( File, Source ) ),
	( file_exists( Source ) ->
		FileLoc = Source
		; 
		bb_get( BBFlag, Dir ),
		atom_concat( '/', File, Trail ),
		atom_concat( Dir, Trail, FileLoc )
	).
