:- use_module( library(lists) ).
:- use_module( library(ugraphs) ).
:- use_module( library(ordsets) ).

bn( X, Y )  :-
	bn( X, [], Y ).
bn([],_All,[]).
bn([RV|RVs], CandParents, BN) :-         %it's a Bayes net if
	bn(RVs,[RV|CandParents],TailBN),                     %this is and ..
	choose_parents(CandParents,Parents),
	list_to_ord_set(Parents,OrdParents),
	write( ord_add_element(TailBN,RV-OrdParents,BN) ), nl,
	ord_add_element(TailBN,RV-OrdParents,BN).
	% write( ord_add_element(TailBN,RV-OrdParents,BN) ), nl.

choose_parents([A,B|CandParents],Parents) :-
	select_parent([A,B|CandParents],OneL,RestCPs),
	select_parent(RestCPs,TwoL,_RemCps),
	ord_union( OneL, TwoL, Parents ).

choose_parents([A|CandParents],ParentL) :-   
	select_parent([A|CandParents],ParentL,_RestCPs).

choose_parents(CandParents,[]).

select_parent(List,OneL,Rest) :-
	length( List, Length ),
	select_parent( Length, List, OneL, Rest ).

select_parent( 7, [H|T], [H], T ).
select_parent( 7, [H|T], Sel, [H|Rest] ) :-
	select_parent( 6, T, Sel, Rest ).
	
select_parent( 6, [H|T], [H], T ).
select_parent( 6, [H|T], Sel, [H|Rest] ) :-
	select_parent( 5, T, Sel, Rest ).
	
select_parent( 5, [H|T], [H], T ).
select_parent( 5, [H|T], Sel, [H|Rest] ) :-
	select_parent( 4, T, Sel, Rest ).

1/4 :: select_parent( 4, [H|T], [H], T ).
3/4 :: select_parent( 4, [H|T], Sel, [H|Rest] ) :-
	select_parent( 3, T, Sel, Rest ).

1/3 :: select_parent( 3, [H|T], [H], T ).
2/3 :: select_parent( 3, [H|T], Sel, [H|Rest] ) :-
	select_parent( 2, T, Sel, Rest ).

1/2 :: select_parent( 2, [H|T], [H], T ).
1/2 :: select_parent( 2, [H|T], Sel, [H|Rest] ) :-
	select_parent( 1, T, Sel, Rest ).

1 :: select_parent( 1, [H], [H], [] ).
1 :: select_parent( 0, _Empty, [], [] ).

