/*
| ?- datime( D1 ), repeat, slp:bn( _Id, _Pin, Path/[], [1,2,3,4,5,6,7,8], BNi ), !, datime(D2).
[1]    Done                          gv -scale 2 submit.ps
D1 = datime(2001,5,23,13,46,48),
D2 = datime(2001,5,23,15,22,1),
BNi = [1-[],2-[1],3-[1,4],4-[6],5-[6],6-[],7-[1,2,3|...],8-[1,2|...]],
Path = [2,11,11,11,11,11,11,11,11,10|...] ?
yes
*/


:- ensure_loaded( library(ordsets) ).	% ord_subtract/3.
:- ensure_loaded( library(ugraphs) ).	% vertices_edges_to_ugraph/3,	
	% add_edges/3, reachable/3.

bn( Nodes, Bn ) :-
	vertices_edges_to_ugraph( Nodes, [], First ),
	First = Aux,
	% vertices_edges_to_ugraph( Nodes, [], Aux ),
	bn( Nodes, Nodes, First, Aux, Bn, _EndAux ).
	% asia_independencies( Bn ).

bn( [], _Nodes, Bn, Aux, Bn, Aux ).
bn( [Vx|Vcs], Nodes, InBn, InAux, OutBn, OutAux ) :-
	bn( Vcs, Nodes, InBn, InAux, MidBn, MidAux ),
	candidate_parents( Vx, Nodes, MidBn, MidAux, CandParents ),
	choose_parents( CandParents, Parents ),
	create_edges( Parents, Vx, ParEdges, ChildEdges ),
	add_edges( MidBn, ParEdges, OutBn ),
	add_edges( MidAux, ChildEdges, OutAux ).

% Vx : the vertex to which we want the candidate parents.
% Nodes : all the nodes of the Bn.
% Aux : the inv(Bn) so far.
% Candidates : Verticies which can become Vx Parents.
% The constraint is that there should be no cycles introduced.
candidate_parents( Vx, Nodes, Bn, Aux, Candidates ) :-
	reachable( Vx, Aux, Reachable ),
	% ord_subtract( Nodes, Reachable, Candidates ).
	ord_subtract( Nodes, Reachable, NoCyclic ),
	donot_introduce_asia_indencictent( NoCyclic, Vx, Bn, Candidates ).

donot_introduce_asia_indencictent( [], _Vx, _Graph, [] ).
donot_introduce_asia_indencictent( [H|T], Vx, Graph, Candidates ) :-
	add_edges( Graph, [Vx-H], TestGraph ),
	asia_indencictent( TestGraph ),
	!,
	Candidates = [H|Tand],
	donot_introduce_asia_indencictent( T, Vx, Graph, Tand ).
donot_introduce_asia_indencictent( [H|T], Vx, Graph, Candidates ) :-
	donot_introduce_asia_indencictent( T, Vx, Graph, Candidates ).

create_edges( [], _Vx, [], [] ).
create_edges( [P|Ps], Vx, [Vx-P|TpEs], [P-Vx|TcEs] ) :-
	create_edges( Ps, Vx, TpEs, TcEs ).

choose_parents([A|CandParents],ParentL) :-   
	select_a_parent([A|CandParents],AL,RestCPs),
	choose_parents( RestCPs, RestL ),
	ord_union( AL, RestL, ParentL ).

choose_parents(_CandParents,[]).

select_a_parent([H|T],[H],T).
select_a_parent([H|T],S,[H|R] ) :-
	select_a_parent(T,S,R).
	/* 
	length( List, Length ),
	select_parent( Length, List, OneL, Rest ).

select_parent( 7, [H|T], [H], T ).
select_parent( 7, [H|T], Sel, [H|Rest] ) :-
	select_parent( 6, T, Sel, Rest ).
	
select_parent( 6, [H|T], [H], T ).
select_parent( 6, [H|T], Sel, [H|Rest] ) :-
	select_parent( 5, T, Sel, Rest ).
	
select_parent( 5, [H|T], [H], T ).
4/5 :: select_parent( 5, [H|T], Sel, [H|Rest] ) :-
	select_parent( 4, T, Sel, Rest ).

1/4 :: select_parent( 4, [H|T], [H], T ).
3/4 :: select_parent( 4, [H|T], Sel, [H|Rest] ) :-
	select_parent( 3, T, Sel, Rest ).

1/3 :: select_parent( 3, [H|T], [H], T ).
2/3 :: select_parent( 3, [H|T], Sel, [H|Rest] ) :-
	select_parent( 2, T, Sel, Rest ).

1/2 :: select_parent( 2, [H|T], [H], T ).
1/2 :: select_parent( 2, [H|T], Sel, [H|Rest] ) :-
	select_parent( 1, T, Sel, Rest ).

1 :: select_parent( 1, [H], [H], [] ).
1 :: select_parent( 0, _Empty, [], [] ).

*/
% :- use_module(library(ugraphs)).
:- use_module(library(lists)).

%Remember that library(ugraphs) treats a-[b] as b being child of a
%neighbours

% asia([1-[],2-[1],3-[2,5],4-[],5-[4],6-[4],7-[3],8-[3,6]]).
% nonasia([1-[4],2-[1],3-[2,5],4-[],5-[4],6-[4],7-[3],8-[3,6]]).
% nonasia2([1-[],2-[1,4],3-[2,5],4-[],5-[4],6-[4],7-[3],8-[3,6]]).


asia_independencies(BN) :-
	vertices(BN,Vs),
	setof(A-B,
		(
		  member(A,Vs),
		  member(B,Vs),
		  A<B,
		  indep(A,B,BN)
		),
		IndependentPairs
	       ),
	IndependentPairs=[1-4,1-5,1-6,2-4,2-5,2-6].

asia_indencictent( Bn ) :-
	asia_indencictent( [1-4,1-5,1-6,2-4,2-5,2-6], Bn ).

asia_indencictent( [], _Bn ).
asia_indencictent( [A-B|T], Bn ) :-
	indep( A, B, Bn ),
	asia_indencictent( T, Bn ).

indep(A,B,BN) :-
	\+ unblocked_path(A,B,BN).

unblocked_path(B,B,_BN).
unblocked_path(A,B,BN) :-
	unblocked_path_via_child(A,B,BN).
unblocked_path(A,B,BN) :-
	unblocked_path_via_parent(A,B,BN).

unblocked_path_via_child(B,B,_BN).	
unblocked_path_via_child(A,B,BN) :-	
	child(A, BN, AChild),
	unblocked_path_via_child(AChild,B,BN).

unblocked_path_via_parent(A,B,BN) :-
	parent(A, BN, AParent),
	unblocked_path(AParent,B,BN).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%utilities

parent(Vertex, [Vertex-Parents|_Rest], Parent) :-
	!,
	member(Parent,Parents).
parent(Vertex, [_Family|Rest], Parent) :-
	parent(Vertex, Rest, Parent).
	
child(Vertex, [Child-Parents|_Rest], Child) :-
	member(Vertex,Parents).
child(Vertex, [_Family|Rest], Child) :-
	child(Vertex, Rest, Child).
