:- use_module(library(ugraphs)).
:- use_module(library(lists)).

%Remember that library(ugraphs) treats a-[b] as b being child of a
%neighbours

asia([1-[],2-[1],3-[2,5],4-[],5-[4],6-[4],7-[3],8-[3,6]]).
nonasia([1-[4],2-[1],3-[2,5],4-[],5-[4],6-[4],7-[3],8-[3,6]]).
nonasia2([1-[],2-[1,4],3-[2,5],4-[],5-[4],6-[4],7-[3],8-[3,6]]).


asia_independencies(BN) :-
	vertices(BN,Vs),
	setof(A-B,
		(
		  member(A,Vs),
		  member(B,Vs),
		  A<B,
		  indep(A,B,BN)
		),
		IndependentPairs
	       ),
	IndependentPairs=[1-4,1-5,1-6,2-4,2-5,2-6].

indep(A,B,BN) :-
	\+ unblocked_path(A,B,BN).

unblocked_path(B,B,_BN).
unblocked_path(A,B,BN) :-
	unblocked_path_via_child(A,B,BN).
unblocked_path(A,B,BN) :-
	unblocked_path_via_parent(A,B,BN).

unblocked_path_via_child(B,B,_BN).	
unblocked_path_via_child(A,B,BN) :-	
	child(A, BN, AChild),
	unblocked_path_via_child(AChild,B,BN).

unblocked_path_via_parent(A,B,BN) :-
	parent(A, BN, AParent),
	unblocked_path(AParent,B,BN).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%utilities

parent(Vertex, [Vertex-Parents|_Rest], Parent) :-
	!,
	member(Parent,Parents).
parent(Vertex, [_Family|Rest], Parent) :-
	parent(Vertex, Rest, Parent).
	
child(Vertex, [Child-Parents|_Rest], Child) :-
	member(Vertex,Parents).
child(Vertex, [_Family|Rest], Child) :-
	child(Vertex, Rest, Child).