/*
From jc@cs.york.ac.uk Mon Mar 12 09:13:59 2001
Date: Sun, 11 Mar 2001 21:41:39 +0000 (GMT)
From: James Cussens <jc@cs.york.ac.uk>
To: nicos@cs.york.ac.uk
Subject: BNs by choosing families


Nicos,

Find below a logic program which constructs BNs by going through each
node and choosing parents for it. To make it an SLP just put numbers
on the decide clause. Note that I have put a limit on the number of
parents. It would be nice to try this with the limit at 2, and see if
we get the Asia network back!

Hasn't been checked that much...

*/

:- use_module(library(ugraphs)).
:- use_module(library(lists)).
:- use_module(library(ordsets)).

/*---------------------------------------------------------------------
  Predicate: bn(+RVs,-BN,-AncBN)
       Args: RVs = a list of random variables
             BN =  a DAG with the RVs as vertices
	     AncBN = a DAG where a node's ancestors in BN are it's parents.
   Succeeds: 2^O(n^2*log n) where n is length of RVs
   Comments: The DAG represents a BN
             The BN need not be fully connected
             AncBN is used for quick checking for cycles
---------------------------------------------------------------------*/

bn(RVs,BN,AncBN) :-
	bn(RVs,RVs,BN,AncBN).

bn([],_AllRVs,[],[]).
bn([RV|RVs],AllRVs,BN,AncBN) :-                         
	bn(RVs,AllRVs,TailBN,AncTailBN),
	choose_parents_no_check(AllRVs,RV,AncTailBN,Parents,RVsAncestors),
	% length(Parents,K),
	% K =< 3,
	upd_anc(TailBN,AncTailBN,RV,RVsAncestors,AncBN1),
	ord_add_element(AncBN1,RV-RVsAncestors,AncBN),
	list_to_ord_set(Parents,OrdParents),
	ord_add_element(TailBN,RV-OrdParents,BN).


upd_anc([],[],_RV,_RVsAncestors,[]).
upd_anc([Node-Family|T],[Node-Ancestors|T2],RV,RVsAncestors,[Node-NewAncestors|TAncBN]) :-
	upd_anc(T,T2,RV,RVsAncestors,TAncBN),
	upd_anc1(Family,Ancestors,RV,RVsAncestors,NewAncestors).

upd_anc1(RVChildsParents,RVChildsAncestors,RV,RVsAncestors,NewAncestors) :-
	member(RV,RVChildsParents),
	ord_union(RVsAncestors,RVChildsAncestors,NewAncestors).
upd_anc1(NotRVChildsParents,NotRVChildsAncestors,RV,_RVsAncestors,NotRVChildsAncestors) :-
	\+ member(RV,NotRVChildsParents).

% Once we see Child, we start checking	
choose_parents_no_check([Child|_T],Child,OldAncBN,Parents,ChildsAncestors) :-
	choose_parents_with_check(OldAncBN,Child,Parents,ChildsAncestors).
choose_parents_no_check([PotParent|T],Child,OldAncBN,Parents,ChildsAncestors) :-
	\+ PotParent=Child,
	choose_parents_no_check(T,Child,OldAncBN,Rest,TChildsAncestors),
	% decide(PotParent,Child,Rest,Parents,Decision),
	decide_ns(Rest,PotParent,Child,Parents,Decision),
	add_ancestor(Decision,PotParent,TChildsAncestors,ChildsAncestors).


%if it's a parent, then it's an ancestor
add_ancestor(yes,Parent,TChildsAncestors,ChildsAncestors) :-
	ord_add_element(TChildsAncestors,Parent,ChildsAncestors).
add_ancestor(no,_NonParent,ChildsAncestors,ChildsAncestors).





choose_parents_with_check([],_Child,[],[]).
choose_parents_with_check([PotParent-PotParentsAncestors|T],Child,ChildsParents,ChildsAncestors) :-
	choose_parents_with_check(T,Child,TChildsParents,TChildsAncestors),
	decide_ns(TChildsParents,PotParent,Child,ChildsParents,Decision),   
	% decide(PotParent,Child,TChildsParents,ChildsParents,Decision),   
			%will need ordering
	childs_ancestors(Decision,PotParentsAncestors,Child,TChildsAncestors,ChildsAncestors).

decide_ns( Rest,PotParent,Child,Parents,Decision) :-
	( Rest = [A,B|T] ->
		Parents = [A,B|T], Decision = no
		;
		decide( PotParent, Child, Rest, Parents, Decision )
	).

decide( NewParent,_Child,Parents,[NewParent|Parents],yes). %include Parent->Child link
decide(_NonParent,_Child,Parents,           Parents , no). %don't include Parent->Child link

/*----------------------------------------------------------------------
a cycle is introduced if
(1) we did decide to add H as a parent of Child, and
(2) Child is one of H's ancestors
----------------------------------------------------------------------*/

%did add a parent
childs_ancestors(yes,ParentsAncestors,Child,TChildAncestors,ChildAncestors) :-
	\+ member(Child,ParentsAncestors),                       %check no cycle introduced
	ord_union(ParentsAncestors,TChildAncestors,ChildAncestors). %only add ancestors if did add parents

%didn't add a parent
childs_ancestors(no,_NonParentAncestors,_Child,TChildAncestors,TChildAncestors).


