:- ensure_loaded( init_lib ). 	% /

:- ensure_loaded( library(select_once_list) ). 	% /3.
:- ensure_loaded( library(head_to_spec) ). 		% /2.
:- ensure_loaded( library(kv_decompose) ). 		% /3.
:- ensure_loaded( library(defined_elsewhere) ). 	% /2.
:- ensure_loaded( library(delete_file_if) ). 	% /2.
:- ensure_loaded( library(lists) ). 		% append/3.
:- ensure_loaded( library(cc) ).			% init_cc/0, next_cc/1, cc/1.
:- ensure_loaded( library(clean_module) ).		% /1.
:- ensure_loaded( library(spec_to_head) ).	%/2.

errhandle(Err)  :- 
	write( errhandle(Err) ), nl.

ad_to_slp( Exclude, TmpFile ) :-
	ad_to_slp( Exclude, delete, TmpFile ).
ad_to_slp( Exclude, Delete, TmpFile ) :-
	clean_module( slp ),
	init_cc,
	% findall( P, predicate_property( ad:P, dynamic ), Ps ),
	ad:bb_get( all_preds, Ps ),
	write( ps(Ps) ), nl,
	select_once_list( Exclude, Ps, Klean ),
	% findall( Spec-L, s_label(Spec,L), SLs ),
	delete_file_if( TmpFile ),
	initial_slp_declarations_to_file( TmpFile ),
	preds_to_file( Klean, TmpFile ),
	compile( TmpFile ),
	pl( sicstus(_S), abolish(ad:_) ),
	% abort,
	( Delete \== dont_delete ->
		% default
		delete_file_if( TmpFile )
		;
		true
	).

preds_to_file( [], _File ).
preds_to_file( [Hspec|Theads], File ) :-
	spec_to_head( Hspec, Hhead ),
	findall( Hhargs-HMB, 
		(ad:clause(Hhead,HBody),Hhead=..[_F|Hhargs], de_module_body(HBody,HMB)),
			Pairs ),
	kv_decompose( Pairs, Heads, Bodies ),
	( is_stochastic(Hhead) -> 
		head_to_spec( Hhead, Hspec ),
		findall( Hl, ad:s_label(Hspec,Hl), HLs ),
		pred_to_s_pred( Hhead, Heads, Bodies, HLs, SPred ),
		record_spred( SPred, File )
		;
		Hhead =.. [F|_OHhargs],
		pred_to_ns_pred( Heads, Bodies, F, NSPred ),
		record_preds( File, NSPred, F )
	),
	preds_to_file( Theads, File ).

pred_to_ns_pred( [], [], _PredName, [] ).
pred_to_ns_pred( [H|Hs], [B|Bs], PredName, [ExpCl|ECls] ) :-
	cc( Id ),
	append( [Id,[Id|Rest]/Lc,[Id|Follows]/Lp], H, ExpList ),
	ExpH =.. [PredName|ExpList],
	( B = true ->
		Rest=Lc,Follows = Lp, ExpCl = ExpH
		;
		expand_body_calls( B, Rest/Lc, Follows/Lp, ExpB ),
		ExpCl = (ExpH:- ExpB)
	),
	next_cc( Id ),
	pred_to_ns_pred( Hs, Bs, PredName, ECls ).

pred_to_s_pred( Proto, Hds, Bds, Lbls, (ExpHead:-ExpBody) ) :-
	Proto =.. [PredName|Args],
	ExpHead =.. [PredName,[H|Pin]/Lc,[Sel|Pout]/Lp|Args],
	cc( CC ),
	TheSelect = select_id( Hds, Args, Lbls, CC, H, Pin, Sel, Sid, Pmid ),
	switch_selector( Hds, Bds, Args, CC, Sid, Pmid/Lc, Pout/Lp, TheSwitch ),
	ExpBody = (TheSelect,TheSwitch).

initial_slp_declarations_to_file( File ) :-
	open( File, append, Stream ),
	write( Stream, ' :- module( slp, [] ).' ),
	nl( Stream ),
	write( Stream, ':- ' ),
	write( Stream, 'compile(library(\'../runtime\')).' ),
	nl( Stream ),
	pl(yap(_Y), (
		write(Stream,':- source.'), nl(Stream) )
		),
	close( Stream ).

% switch_selector( [], _Bodies, _Labels, Count, Id, TheSwitch ) :-
switch_selector( [HArgs], [Bd], Args, CC, _Id, Pin/Lc, Pout/Lp, Switch ) :-
	!,
	next_cc( CC ),
	expand_body_calls( Bd, Pin/Lc, Pout/Lp, ExpBd ),
	Switch = (HArgs=Args,ExpBd).
switch_selector( [HArgs|Thas], [Bd|Bds], Args, CC, Id, Pin, Pout, Switch ) :-
	expand_body_calls( Bd, Pin, Pout, ExpBd ),
	Switch = ( (Id=:=CC-> HArgs=Args, ExpBd ; NestSwitch) ),
	NxtCC is CC + 1,
	switch_selector( Thas, Bds, Args, NxtCC, Id, Pin, Pout, NestSwitch ).

expand_body_calls( (A->B;C), Pathin/Ld, Pathout/Lp, ExpIf ) :-
	!,
	ExpIf = ( ExpA->((ExpB,Ld=Le,Lp=LpB);(ExpC,Ld=Lf,Lp=LpC)) ),
	expand_body_calls( A,  Pathin/LcA, Pathout/LpA, ExpA ),
	expand_body_calls( B, LcA/Le, LpA/LpB, ExpB ),
	expand_body_calls( C, LcA/Lf, LpA/LpC, ExpC ).
expand_body_calls( (A,B), Pathin/Lc, Pathout/Lp, (ExpA,ExpB) ) :-
	!,
	expand_body_calls( A,  Pathin/LcA, Pathout/LpA, ExpA ),
	expand_body_calls( B, LcA/Lc, LpA/Lp, ExpB ).
expand_body_calls( (\+ PrpA), Pathin/Lc, Pathout/Lp, (\+ ExpA) ) :-
	!,
	meta_ad_strip( PrpA, A ),
	expand_body_calls( A, _Fin/_FLcA, _Fout/_FLpA, ExpA ),
	Pathin = Lc, Pathout = Lp.
expand_body_calls( findall(A,PrpB,C), Pathin/Lc, Pathout/Lp, ExpF ) :-
	!,
	meta_ad_strip( PrpB, B ),
	expand_body_calls( B,  _Fin/_FLcA, _Fout/_FLpA, ExpB ),
	Pathin = Lc, Pathout = Lp,
	ExpF = findall(A,ExpB,C).
expand_body_calls( setof(A,PrpB,C), Pathin/Lc, Pathout/Lp, ExpF ) :-
	!,
	meta_ad_strip( PrpB, B ),
	expand_body_calls( B,  _Fin/_FLcA, _Fout/_FLpA, ExpB ),
	Pathin = Lc, Pathout = Lp,
	ExpF = setof(A,ExpB,C).
expand_body_calls( PrpA, Pathin/Lc, Pathout/Lp, ExpA ) :-
	meta_ad_strip( PrpA, A ),
	( is_stochastic(A) ->
		A =.. [PredName|Args],
		ExpA =.. [PredName,Pathin/Lc,Pathout/Lp|Args]
		;
		( defined_elsewhere(A,Mod) -> 
			% Pathin = Lc,
			% Pathout = Lp,
			ExpA = (Mod:A,Pathin = Lc,Pathout=Lp)
			% ExpA = (Mod:A, Pathout = Pathin)
			;
			A =.. [PredName|Args],
			append( [PredName,_,Pathin/Lc,Pathout/Lp], Args, ExpAList ),
			ExpA =.. ExpAList
		)
	).

is_stochastic( Head ) :-
	head_to_spec( Head, Spec ),
	ad:pred_type(Spec, s).

record_spred( SPred, File ) :-
	open( File, append, Stream ), 
	current_output( Output ),
	set_output( Stream ),
	portray_clause( SPred ), 
	set_output( Output ),
	close( Stream ).

record_preds( File, Pairs, PredName ) :-
	open( File, append, Stream ), 
	current_output( Output ),
	set_output( Stream ),
	record_ns_clauses( Pairs, PredName ),
	set_output( Output ),
	close( Stream ).

record_ns_clauses( [], _PredName ).
record_ns_clauses( [Clause|T], PredName ) :-
	portray_clause( Clause ), 
	record_ns_clauses( T, PredName ).

de_module_body( (A,B), (DemA,DemB) ) :-
	!,
	de_module_body( A, DemA ),
	de_module_body( B, DemB ).
de_module_body( (A->B;C), (DemA->DemB;DemC) ) :-
	!,
	de_module_body( A, DemA ),
	de_module_body( B, DemB ),
	de_module_body( C, DemC ).
de_module_body( (A;B), (DemA;DemB) ) :-
	!,
	de_module_body( A, DemA ),
	de_module_body( B, DemB ).
de_module_body( _Mod:A, A ) :- !.
de_module_body( A, A ).

meta_ad_strip( PrpA, A ) :-
	( PrpA = ad:A -> 
		true
		; 
		A = PrpA
	).
