/* 
From jc@cs.york.ac.uk Wed Feb 21 15:21:09 2001
Date: Wed, 21 Feb 2001 13:12:18 +0000 (GMT)
From: James Cussens <jc@cs.york.ac.uk>
To: nicos@cs.york.ac.uk
Subject: code of computing ratios of likelihoods


Nicos,

Here it is. I haven't littered it with cuts. As long as Sicstus is
behaving it should realise that the code is deterministic.


James

*/

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% interpret this first one as a BN where 1 and 5 are the parents of 3
% eg Node-[Parent1,Parent2]

:- use_module(library(ugraphs)).
:- use_module(library(lists)).
:- use_module(library(terms)).
:- use_module(library(ordsets)).

:- dynamic cached_family_contrib/3.

%BNs look like this

bn(1,[1-[5],2-[],3-[1,5],4-[2],5-[4]]).
bn(2,[1-[5],2-[1,5],3-[4],4-[1],5-[]]).
bn(3,[1-[2,5],2-[5],3-[1,4,5],4-[1,5],5-[]]).

%data looks like this
%have all binary 

/* 
data(0,0,1,1,0).
data(0,1,1,1,0).
data(0,0,0,1,0).
data(1,0,1,1,0).
data(0,1,1,1,0).
data(0,0,1,1,0).
data(0,0,1,0,0).
data(0,1,1,1,0).
data(0,0,0,2,0).
data(1,0,1,1,0).
data(0,1,1,4,0).
data(0,0,1,1,0).

*/

:- 	user:datafile( Data ),
	user:write( using_data(Data) ), nl,
	compile( Data ).

% either grab cached value or compute it

lhood_family_contrib(Family,Contrib) :-
	term_hash(Family,HashValue),                
	(
	    cached_family_contrib(HashValue,Family,Contrib) ->
	    
	    true ;
	    
	    compute_lhood_family_contrib(Family,Contrib),
	    assert(cached_family_contrib(HashValue,Family,Contrib))
	).

% eventually this might call an external function

compute_lhood_family_contrib(Family,Contrib) :-
	construct_goal(Family,Term,Goal),
	findall(Term,Goal,Instances),                   % constructs massive list
	instances_to_counts(Instances,ChildCountsList),
	write(ChildCountsList),nl,
	product_of_multinomials(ChildCountsList,Contrib).


construct_goal(Child-Parents,PVars-ChildVar,Goal) :-
	Goal = data(_,_,_,_,_,_,_,_),
	bind([Child|Parents],Goal,[ChildVar|PVars]).

% yanks out relevant variables from Goal and puts them in 3rd argument
bind([],_,[]).
bind([H|T],Goal,[H1|T1]) :-
	arg(H,Goal,H1),
	bind(T,Goal,T1).


instances_to_counts(Instances,ChildCountsList) :-
	keysort(Instances,SortedInstances),          %order by values of parents
	do(SortedInstances,[],ChildCountsList).      %plod through making a list of lists

do([],ChildCountsList,ChildCountsList).
do([PVals-ChildVal|T],SoFar,ChildCountsList) :-
	child_counts([PVals-ChildVal|T],PVals,[],ChildCounts,Rest),
	do(Rest,[ChildCounts|SoFar],ChildCountsList).


% get counts for the various observed values of a child
% for a particular parent instantiation

child_counts([PVals-ChildVal|T],PVals,Sofar,Out,Rest) :-        %got a PVals match
	!,
	(
	    select(ChildVal-Count,Sofar,Residue) ->             %if seen this ChildVal before
	    C is Count + 1,                                     %increment count
	    NewSofar = [ChildVal-C|Residue] ;                   %put back
	    NewSofar = [ChildVal-1|Sofar]                       %else start the count for this ChildVal
	),
	child_counts(T,PVals,NewSofar,Out,Rest).
child_counts(L,_PVals,Out,Out,L).                               %match failed, we're done with this PVals

strip_vals(In,Out) :-
	strip_vals(In,[],Out).

strip_vals([],Out,Out).
strip_vals([_Val-H|T],In,Out) :-
	strip_vals(T,[H|In],Out).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% START NUMERICAL UTILITIES 
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% no choice points created here (hopefully!)

% computes multinomial

product_of_multinomials(List,Prod) :-
	product_of_multinomials(List,1,Prod).

product_of_multinomials([],Out,Out).
product_of_multinomials([H|T],In,Out) :-
	multinomial(H,F),
	Mid is In * F,
	product_of_multinomials(T,Mid,Out).

multinomial(Numbers,Multinomial_Coefficient) :-
	strip_vals(Numbers,Numbers1),            % end up with pure numbers
	product_of_factorials(Numbers1,1,Prod),
	sum_list(Numbers1,Sum),                  % library predicate
	factorial(Sum,Sum_Factorial),
	Multinomial_Coefficient is Prod/Sum_Factorial.

product_of_factorials([],Out,Out).
product_of_factorials([H|T],In,Out) :- %not currently using Value
	factorial(H,F),
	Mid is In * F,
	product_of_factorials(T,Mid,Out).

% computes factorial

factorial(N,Fact) :-
	factorial(N,1,Fact).

factorial(0,Out,Out) :- !.
factorial(N,In,Out) :-
	Mid is N * In,
	NN is N - 1,
	factorial(NN,Mid,Out).
	
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% END NUMERICAL UTILITIES 
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% START COMPUTATION OF LIKELIHOOD
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

lhood(BN,L) :-
	lhood(BN,1,L).

% for each family
% does outermost loop of Heckerman's (32)

lhood([],L,L).
lhood([Family|Families],In,Out) :-
	lhood_family_contrib(Family,Family_Contrib),   %this is the 'score' for this family
	Mid is In * Family_Contrib,
	lhood(Families,Mid,Out).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% END COMPUTATION OF LIKELIHOOD
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% START COMPUTATION OF LIKELIHOOD RATIO
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

lhood_ratio(BN1,BN2,Ratio) :-
	lhood_ratio(BN1,BN2,1,Ratio).

% OK, done!
lhood_ratio([],[],L,L).                 

% if families are identical then both cancel out, and can be ignored
lhood_ratio([Child-Parents|Rest1],[Child-Parents|Rest2],In,Out) :-
	!,
	lhood_ratio(Rest1,Rest2,In,Out).

% OK, there are different parents for Child, so have to compute ratio, sigh...
lhood_ratio([Family1|Rest1],[Family2|Rest2],In,Out) :-
	lhood_family_contrib(Family1,Family_Contrib1),
	lhood_family_contrib(Family2,Family_Contrib2),
	% write( Mid is In * (Family_Contrib1/Family_Contrib2) ), nl,
	Mid is In * (Family_Contrib1/Family_Contrib2),
	lhood_ratio(Rest1,Rest2,Mid,Out).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% END COMPUTATION OF LIKELIHOOD RATIO
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

