% :- module( kernel_vp, [kernel_vp/4] ).

:- ensure_loaded( library(random) ).			%/1.
:- ensure_loaded( library(lists) ).
:- ensure_loaded( jump_bn ).

kernel( Nodes, Iter, P, VorO, Model ) :-
	repeat,
	slp:bn( _Id, _Pin, Path/[], Nodes, BNi ),
	!,
	user:dbg( 28, c(BNi) ),
	path_to_btrack_pos( Path, 1, Pos ),
	length( Pos, Ni ),
	reverse( Pos, Sop ),
	(VorO == op -> BpStr = P ; BpStr = 1-1/2-P), 
	transition( Path, Nodes, Sop, BNi, Ni, BpStr, 1, [iter(Iter)], Model ).

% transition( [], Metrics, _Bp, _Termin, Metrics ).
transition( InPath, Nds, Sop, BNi, Ni, BpStr, Iter, Termin, Model ) :-
	% dbg( 20, inpath__befire_backtrack(InPath) ),
	( BpStr = Pth-Bp-Pl -> true ; Bp = BpStr ),
	backtrack_on_nxt( InPath, Nds, Sop, Bp, LCi, LCst, NwPath, BNst ),
	% p_limit( Pl ),
	( BpStr = Pth-Bp-Pl ->
		( Pth =:= Pl ->
			NxBpStr = 1-1/2-Pl
			;
			NxPth is Pth + 1,
			Denom is 2 ** NxPth,
			NxBp is (Denom - 1) / Denom,
			NxBpStr = NxPth-NxBp-Pl
		)
		;
		NxBpStr = BpStr
	),
	NxIter is Iter + 1, % changed on 2003-09-05.
	( LCst == f ->
		% dbg( 21, failed_backtracking ),
		% NxIter is Iter,
		OutPath = InPath, NxNi is Ni,
		NxBNi = BNi, OutSop = Sop,
		display_bns_in_iteration( Iter, BNi, ['algorithm_hit_a_failure_branch'-[]], none, dont_jump )
		;
		% NxIter is Iter + 1,
		% dbg( 23, nwpath(NwPath) ),
		user:dbg( 24, b(BNst) ),
		path_to_btrack_pos( NwPath, 1, NwPos ),
		reverse( NwPos, NwSop ),
		length( NwPos, Nst ),
		jump_bn( BNi, BNst, Ni, Nst, Iter, LCi, LCst, Bp, Jump ),
		( Jump==jump ->
			% dbg( 27, jumping ),
			OutPath = NwPath, NxNi is Nst,
			NxBNi = BNst, OutSop = NwSop
			; 
			% dbg( 27, not_jumping ),
			OutPath = InPath, NxNi is Ni,
			NxBNi = BNi, OutSop = Sop
		)
	),
	user:dbg( 28, c(NxBNi) ),
	( termination( Termin, Jump, Iter ) ->
		% dbg( 28, c(NxBNi) ),
		Model = Jump-OutPath
		;
		% write( nxp(NxP) ), nl,
		transition( OutPath, Nds, OutSop, NxBNi, NxNi, NxBpStr, NxIter, Termin, Model )
	).

backtrack_on_nxt( InPath, Nds, Sop, Bp, LblCi, LblCst, NewPath, BNst ) :-
	( Sop == [] ->
		write( cannot_backtrack_on_empty_choice_points ), nl,
		abort
		;
		select_btrack_point( Sop, Bp, Nth ),
		% dbg( 28, select_btrack_point( Sop, Bp, Nth ) ),
		split_on_nth( Nth, InPath, Left, _Right ),
		% dbg( 28, split_on_nth( Nth, InPath, Left, _Right ) ),
		% Nds = [1,2,3,4,5,6,7,8],
		( slp:bn(_Sd,Left/_Sin,PrvPath/[],Nds,BNst) -> 
		% ( slp:bn(_Sd,Left/_Sin,NewPath/[],Nds,BNst) -> 
		% ( slp:bn(_Sd,Left/_Sin,NewPath/[],Nds,BNst,_At) -> 
			% dbg_list( 28, inpath, InPath ),
			% dbg_list( 28, prvpath, PrvPath ),
			% dbg( 28, nth(Nth) ),
			%%%% INVESTIGATE THIS ...
			% nth( Nth, InPath, _Sin/LblCi ),
			% nth_replace_column( Nth, PrvPath, LblCst, NewPath )
			%%% ALTERNATIVELY (FROM KERNEL_OP, 2003-12-02)
			% ( slp:bn(_Sd,Left/_Sin,NewPath/[],Nds,BNst) -> 
			nth( Nth, InPath, _Sin/LblCi ),
			nth( Nth, InPath, _Sou/LblCst ),
			NewPath = PrvPath
			%%%

			% nth( Nth, PrvPath, NewPPoint ),
			% dbg( 28, newpath(NewPath) ),
			% dbg( 28, lbs(old(LblCi)-new(LblCst)) )

			% nth( Nth, InPath, _Sou/LblCst )
			% 21st May 2001
			% nth( Nth, NewPath, _Sou/LblCst )
			% dbg( 29, successfull_bactracking )
			;
			% dbg( 28, bn_call_failed ),
			% write( inpath(InPath) ), nl,
			% write( nth(Nth) ), nl,
			% write( left(Left) ), nl,
			LblCst = f,
			NewPath = [], BNst = []
			% dbg( 29, failed_bactracking )
		)
	).

path_to_btrack_pos( [], _Ind, [] ).
path_to_btrack_pos( [H|T], Ind, Pos ) :-
	( H=_Id/_Prb ->
		Pos = [Ind|TPos]
		;
		Pos = TPos
	),
	NxInd is Ind + 1,
	path_to_btrack_pos( T, NxInd, TPos ).

select_btrack_point( [H|T], P, Nth ) :-
	( T == [] ->
		% dbg( 30, controlled_sample_from_last(H) ),
		Nth = H
		;
		( stop_backtracking(P) ->
			Nth = H
			;
			select_btrack_point( T, P, Nth )
		)
	).

% nth_replace_column( Nth, InPath, LblCst, NewPath ) :-
nth_replace_column( Nth, [H|T], LblCst, NewPath ) :-
	( Nth =< 1 -> 
		( H = Sid:LblCst -> 
			NewPath = [Sid|T]
			;
			H = Sid/LblCst,
			NewPath = [H|T]
		)
		;
		NxN is Nth - 1,
		NewPath = [H|NewT],
		nth_replace_column( NxN, T, LblCst, NewT )
	).

split_on_nth( Nth, [H|T], [H|Left], Right ) :-
	( Nth =:= 1 -> 
		Left = [],
		Right = T
		;
		NxNth is Nth - 1,
		split_on_nth( NxNth, T, Left, Right )
	).

termination( Termin, _Jump, Iter ) :-
	( memberchk( iter(IterLim), Termin ) ->
		Iter >= IterLim, user:dbg( 81, ' TERMINATING ... ' )
		;
		user:dbg( 82, iteration(Iter) ),
		fail
	).

stop_backtracking( Bp ) :-
	random( Rnd ),
	Rnd > Bp.
