:- module( kernel_op, [kernel_op/4] ).

:- ensure_loaded( library(random) ).			%/1.
:- ensure_loaded( library(lists) ).
:- ensure_loaded( jump_bn ).

kernel_op( Nodes, Iter, P, Model ) :-
	repeat,
	slp:bn( _Id, _Pin, Path/[], Nodes, BNi ),
	!,
	user:dbg( 28, c(BNi) ),
	path_to_btrack_pos( Path, 1, Pos ),
	length( Pos, Ni ),
	reverse( Pos, Sop ),
	% write( first_number_of_bactrack_points(Ni) ), nl,
	transition( Path, Nodes, Sop, BNi, Ni, P, 1, [iter(Iter)], Model ).

% transition( [], Metrics, _Bp, _Termin, Metrics ).
transition( InPath, Nds, Sop, BNi, Ni, Bp, Iter, Termin, Model ) :-
	% dbg( 20, inpath__befire_backtrack(InPath) ),
	backtrack_on_nxt( InPath, Nds, Sop, Bp, LCi, LCst, NwPath, BNst ),
	NxIter is Iter + 1, % changed on 2003-09-05.
	( LCst == f ->
		% dbg( 21, failed_backtracking ),
		% NxIter is Iter, 
		OutPath = InPath, NxNi is Ni,
		NxBNi = BNi, OutSop = Sop,
		display_bns_in_iteration( Iter, BNi, ['algorithm_hit_a_failure_branch'-[]], none, dont_jump )
		;
		% NxIter is Iter + 1,
		% dbg( 23, nwpath(NwPath) ),
		user:dbg( 24, b(BNst) ),
		path_to_btrack_pos( NwPath, 1, NwPos ),
		% length( NwPos, NofBps ),
		% write( number_of_bactrack_points(NofBps) ), nl,
		reverse( NwPos, NwSop ),
		length( NwPos, Nst ),
		jump_bn( BNi, BNst, Ni, Nst, Iter, LCi, LCst, Bp, Jump ),
		( Jump==jump ->
			% dbg( 27, jumping ),
			OutPath = NwPath, NxNi is Nst,
			NxBNi = BNst, OutSop = NwSop
			; 
			% dbg( 27, not_jumping ),
			OutPath = InPath, NxNi is Ni,
			NxBNi = BNi, OutSop = Sop
		),
		user:dbg( 28, c(NxBNi) )
	),
	( termination( Termin, Jump, Iter ) ->
		% dbg( 28, c(NxBNi) ),
		Model = Jump-OutPath
		;
		transition( OutPath, Nds, OutSop, NxBNi, NxNi, Bp, NxIter, Termin, Model )
	).

backtrack_on_nxt( InPath, Nds, Sop, Bp, LblCi, LblCst, NewPath, BNst ) :-
	( Sop == [] ->
		write( cannot_backtrack_on_empty_choice_points ), nl,
		abort
		;
		select_btrack_point( Sop, Bp, Nth ),
		split_on_nth( Nth, InPath, Left, _Right ),
		% Nds = [1,2,3,4,5,6,7,8],
		( slp:bn(_Sd,Left/_Sin,NewPath/[],Nds,BNst) -> 
			nth( Nth, InPath, _Sin/LblCi ),
			nth( Nth, InPath, _Sou/LblCst ),
			user:dbg( 29, successfull_bactracking  )
			;
			LblCst = f,
			NewPath = [], BNst = [],
			user:dbg( 29, failed_bactracking )
		)
	).

path_to_btrack_pos( [], _Ind, [] ).
path_to_btrack_pos( [H|T], Ind, Pos ) :-
	( H=_Id/_Prb ->
		Pos = [Ind|TPos]
		;
		Pos = TPos
	),
	NxInd is Ind + 1,
	path_to_btrack_pos( T, NxInd, TPos ).

select_btrack_point( [H|T], P, Nth ) :-
	( T == [] ->
		% dbg( 30, controlled_sample_from_last(H) ),
		Nth = H
		;
		( stop_backtracking(P) ->
			% dbg( 34, controlled_sample_from(H) ),
			Nth = H
			;
			select_btrack_point( T, P, Nth )
		)
	).

split_on_nth( Nth, [H|T], [H|Left], Right ) :-
	( Nth =:= 1 -> 
		Left = [],
		Right = T
		;
		NxNth is Nth - 1,
		split_on_nth( NxNth, T, Left, Right )
	).
	
termination( Termin, _Jump, Iter ) :-
	( memberchk( iter(IterLim), Termin ) ->
		Iter > IterLim, user:dbg( 81, ' TERMINATING ... ' )
		;
		user:dbg( 82, iteration(Iter) ),
		fail
	).

stop_backtracking( Bp ) :-
	random( Rnd ),
	Rnd > Bp.
