:- ensure_loaded( library(system) ).  	% file_exists/1.
:- ensure_loaded( library(lists) ).  	% member/2, append/3.
:- ensure_loaded( flatten ).
:- ensure_loaded( break_list_on_list ).

unique_filename( FileAtom, UniqueFile ) :-
	(file_exists( FileAtom ) ->
		atom_codes( FileAtom, FileString ),
		get_next_filename( FileString, NextFileString ),
		atom_codes( NextFileAtom, NextFileString ),
		print_message( warning, 
			file_name_change_due_to_existance(FileAtom, NextFileAtom) ),
		unique_filename( NextFileAtom, UniqueFile )
		;
		UniqueFile = FileAtom
	),
	!.

get_next_filename( CharListIn, CharListOu_T ) :-
     ( (break_list_on_list( CharListIn, ".", Main, Exts ),
        \+ member( 0'/, Exts ) )
          -> append( ".", Exts, DotExts )
		; 
          ( append( Main, "/", CharListIn ) ->
               DotExts = "/"
               ;
               Main = CharListIn, DotExts = []
          )
     ),
     ( break_list_on_list( Main, "_vu", Trunk, Vers )
          ->  get_next_version( Vers, NextVers )
          ;   Trunk = Main, NextVers = "01"
     ),
     flatten( [Trunk,"_vu",NextVers,DotExts], CharListOu_T ).

get_next_version( [F,S], [F1,S1] ) :-
         ( S<57  ->
             S1 is S + 1
             ;
             ( S=57 -> S1 is 97
                     ;
               ( S < 122 ->
                    S1 is S + 1
                    ;
                    S1 is 48, F1 is F + 1
               )
             )
           ),
           ( var(F1) -> F1 is F ; true ).

